/*
 *   devil.c         chat server che si appoggia all'inetd per
 *   Vers. 0.9.5   l'aggancio delle connessioni telnet client.           
 *                 Copyright (C)  1999-2002  Matteo Nastasi 
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *     CONTACT:    nastasi@alternativeoutput.it
 *                 matteo.nastasi@milug.org
 *
 */


#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <unistd.h>
#include <crypt.h>
#include <time.h>

/* path radice */
#define MAINPATH "/usr/tmp"
#define PREFIX "/usr/tmp/devil"
/* timeout di inattivita' globale della chat */
#define TIMEOUT 120000L
/* file con le password */
#define PASSWDFILE "/etc/devilpasswd"
/* timeout per l'inserimento di caratteri */
#define TIMEOINS 200L
/* numero di righe nell'history */
#define N_RIGHE 20

#define DEBUG
#define DEBUG3 

void display_history (char **,int);
void display_help (void);
int telfgets (char *,int,int,int,long,int,int);
int concurr_append (char *,char *);
int filelock(int,int,long,long,int);
long check_users(char *,long,long *);
int remove_users(long);
int display_users(void);
char *cleanstr(char *);
void myexit (int,char *);
char *sprintip(char *s, unsigned long *laddr);
int check_account(char *user, char *pass);

long rpid = -1,lpid = -1;
char *bigsp = "                                                              ";
FILE *fplog = NULL;
#ifdef DEBUG
  FILE *fpdbg = NULL;
#endif

unsigned long trust_ip[] = { 0x0100007f, 
			     0x000116ac, 
			     00 };
unsigned long trust_msk[]= { 0xffffffff,
			     0x00ffffff,
			     00 };

/* file storico per la chat */
/* file dove vengono aggiornati gli utenti connessi */
/* file di log per la chat */
char history[255], fusers[255], devlog[255], debug[255];

char *sprintip(char *s, unsigned long *laddr)
{
  char *p;
  
  p = (char *)laddr;

  sprintf(s,"%d.%d.%d.%d",p[0], p[1], p[2], p[3]);

  return (s);
}
char *cleanstr(char *s)
{
  int i,e;
  
  for (i=0 , e=0 ; (i == 0 || s[i-1] != '\0') ; i++)
    if (isprint(s[i]) || s[i] == '\a')
      s[e++] = s[i];
  return s;
}

int check_account(char *user, char *pass)
{
  char bf[512], *hash, *ret;
  FILE *fp;
  
  if (user[strlen(user)-1] == '\n')
    user[strlen(user)-1] = '\0';

  if (pass[strlen(pass)-1] == '\n')
    pass[strlen(pass)-1] = '\0';

  if ((fp = fopen(PASSWDFILE, "r")) == NULL) 
    return (0);

  while (fgets(bf,512,fp)) {
    if (!strncmp(user,bf,strlen(user))) {
      /* user trovato, ora confronta la password */
      hash = &bf[strlen(user)+1];
      if (hash[strlen(hash)-1] == '\n')
	hash[strlen(hash)-1] = '\0';
      ret = crypt(pass,hash);
      if (!strcmp(ret,hash)) {
	fclose(fp);
	return (1);
      }
      else 
	break;
    }
  }

  fclose(fp);
  return (0);
}
    

int main(int argc,char **argv)
{
  FILE *fp;
  int i,nl,ret,sret;
  char nome[9],pass[9], line[80],lbf[N_RIGHE][80],*pbf[N_RIGHE],bf[1024],bf2[1024],sbf[1024];
  fd_set fd;
  struct timeval timeo;
  long ct;
  long pid;
  unsigned long laddr;
  struct sockaddr saddr;
  struct sockaddr_in *saddr_in;
  socklen_t lsaddr;
  char *suffix;

  suffix = argv[1];
  lsaddr = sizeof(struct sockaddr);

  
  sprintf(history,"%s_%s.txt",PREFIX,suffix);
  sprintf(fusers,"%s_%s.usr",PREFIX,suffix);
  sprintf(devlog,"%s_%s.log",PREFIX,suffix);
  sprintf(debug,"%sdebug_%s.log",PREFIX,suffix);
  
  for (i=0 ; i<N_RIGHE ; i++)
    pbf[i] = lbf[i];
  strcpy(nome,"UNKNOW");
  pid = getpid();
#ifdef DEBUG
  if ((fpdbg = fopen(debug,"a+")) == NULL) {
    myexit(1,NULL);
  }
#endif
  if ((fplog = fopen(devlog,"a+")) == NULL) {
    myexit(2,NULL);
  }
  fprintf(fplog,"%s %ld\n",argv[0],pid);
  
  
  if ((fp = fopen(history,"r+")) == NULL)
    if ((fp = fopen(history,"w+")) == NULL)
      myexit(3,NULL);

  for (nl = 0 ; fgets(lbf[nl%20],80,fp) != NULL ; nl++);
  
  if (getsockname(0,&saddr,&lsaddr) == -1)
    myexit(8,NULL);
  
  if (saddr.sa_family == AF_INET) {
    saddr_in = (struct sockaddr_in *)&saddr;
    laddr = *((unsigned long *)&saddr_in->sin_addr);
  }
#ifndef TELRAW
  fprintf(stdout,"%c%c%c%c%c%c",255,251,3,255,251,1);
#endif

  /* verifica se appartenente a una rete degna di fiducia */
  for (i = 0 ; trust_ip[i] != 0 ; i++) {
    if ((laddr & trust_msk[i]) == (trust_ip[i] & trust_msk[i]))
      break;
  }
  if (trust_ip[i] == 0) {
    sprintf(bf,"\r\nAccesso da indirizzo sconosciuto (%s), user e password richiesti\r\n", 
	    sprintip(bf2, &laddr));
    write(1,bf,strlen(bf));
    /* MEMO qui check di user e pass */

    for (i = 0 ; i < 3 ; i++) {
      fprintf(stdout,"\r\nUser: ");
      fflush(stdout);
      if ((ret = telfgets(nome,8,0,1,TIMEOINS,0,0)) != 1)
	myexit(4,NULL); 
      fprintf(stdout,"\r\nPassword: ");
      fflush(stdout);
      if ((ret = telfgets(pass,8,0,1,TIMEOINS,0,1)) != 1)
	myexit(4,NULL); 
      if (check_account(nome,pass))
	break;
      sleep(2);
      fprintf(stdout,"\r\nLo user e/o la password sono sconosciute, riprovare\r\n");
      fflush(stdout);
    }
    if (i == 3)
      myexit(7,NULL);
  }


  fprintf(stdout,"\r\nBenvenuti nella DEVIL CHAT\r\n");

  for (i = 0 ; i < 3 ; i++) {
    fprintf(stdout,"\r\nDigita il tuo nickname (max 8 char alfanumerici): ");
    fflush(stdout);
    if ((ret = telfgets(nome,8,0,1,TIMEOINS,0,0)) != 1)
      myexit(4,nome); 
    if ((rpid = check_users(nome,pid,&lpid)) > 0L && strlen(nome) > 0)
      break;
    else if (rpid == -1L)
      myexit(35,"UNKNOW");
    fprintf(stdout,"\r\nIl nickname e' gia` usato da un'altro utente "
	           "connesso\r\noppure hai usato caratteri non premessi\r\n");
    fflush(stdout);
  }
  if (i == 3)
    myexit(7,nome);

  write(1,"\r\n\r\n",4);
  display_history(pbf,nl);
  
  sprintf(sbf,"Ingresso di %s\r\n",nome);
  if (concurr_append(history,sbf) < 0)
    myexit(5,nome);
  while(ct < TIMEOUT) {
#ifdef DEBUG2
    fprintf(fpdbg,"while %d\n",ct);
    fflush(fpdbg);
#endif
    while (fgets(lbf[nl%20],80,fp) != NULL) {      
      lbf[nl%20][strlen(lbf[nl%20]) - 1] = '\0';
      sprintf(sbf,"%s\r\n",lbf[nl%20]);
      write(1,sbf,strlen(sbf));
      nl++;
      ct = 0;
    } 
    FD_SET(0,&fd);
    timeo.tv_sec  = 0L;
    timeo.tv_usec = 100000L;
    if ((sret = select(0+1, &fd, NULL, NULL, &timeo)) > 0) {
      if (FD_ISSET(0,&fd)) {
	sprintf(sbf,"RIGA: ");
	write(1,sbf,strlen(sbf));
	if ((ret = telfgets(line,54,0,1,TIMEOINS,1,0)) == -1) {
	  /* connessione chiusa */
	  myexit(0,nome);
	}
	else if (ret == 0) {
	  /* timeout */
	  sprintf(sbf,"\r%s\r",bigsp);
	  write(1,sbf,strlen(sbf));	    
	}
	else {
	  cleanstr(line);
	  /* caso in cui e' stato scritto qualcosa */
	  if (!strcmp(line,"end") || !strcmp(line,"quit")) {
	    sprintf(bf,"\r\n\r\nAlla prossima %s%s",nome,bigsp);
	    bf[77] = '\0';
	    sprintf(sbf,"%s\r\n\r\n",bf);
	    write(1,sbf,strlen(sbf));
	    myexit(0,nome);
	  }
	  else if (!strcmp(line,"hist")) {
	    display_history(pbf,nl);
	  }
	  else if (!strcmp(line,"help")) {
	    display_help();
	  }
	  else if (!strcmp(line,"who")) {
	    display_users();
	  }
	  else {
	    char tbf[64];
	    time_t t;
	    time(&t);
	    strftime(tbf,64,"%d/%H.%M",localtime(&t));
 	    sprintf(bf,"%s:%8s: %s%s",tbf,nome,line,bigsp);
	    bf[77] = '\0';
	    sprintf(sbf,"%s\r\n",bf);
	    if (concurr_append(history,sbf) < 0)
	      myexit(6,nome);
            if (line[0] == '#') {
                FILE *bcfp;
                char bcbf[512];
                char bcbf2[512];
                char *acapo;

                sprintf(bcbf,"echo '%s'| bc -lq\n", &line[1]);
                
                if ((bcfp = popen(bcbf, "r")) != NULL) {
                    while (fgets(bcbf,512,bcfp)) {
                        if ((acapo = strchr(bcbf,'\n')) != NULL)
                            *acapo = '\0';
                        sprintf(bcbf2, " = %s\r\n", bcbf);
                        if (concurr_append(history,bcbf2) < 0)
                            myexit(7,nome);
                    }
                    pclose(bcfp);
                }
            }
	  }
	}
      }
      ct = 0;
    }
    else {
#ifdef DEBUG2
      fprintf(fpdbg,"sret da main %d\n",sret);
      fflush(fpdbg);
#endif
      ct++;    
    }
  } 
  myexit(0,"fine");

  return (0);
}


void display_history(char **lbf,int nl)
{
  int i,l;
  char sbf[1024];
  
  l = (nl > N_RIGHE ? nl - N_RIGHE : 0);      
  for (i = l ; i<nl ; i++) {
    lbf[i%N_RIGHE][strlen(lbf[i%N_RIGHE]) - 1] = '\0';
    sprintf(sbf,"%s\r\n",lbf[i%N_RIGHE]);
    write(1,sbf,strlen(sbf));
  }
}

void display_help() {
  char sbf[1024];
  
    sprintf(sbf,"\r\n---- H E L P -------------------\r\n");
       write(1,sbf,strlen(sbf));
    sprintf(sbf," who    chi e' connesso\r\n");
       write(1,sbf,strlen(sbf));
    sprintf(sbf," hist   gli ultimi %d messaggi\r\n",N_RIGHE);
       write(1,sbf,strlen(sbf));
    sprintf(sbf," quit   uscita dalla chat\r\n");
       write(1,sbf,strlen(sbf));
    sprintf(sbf," # <expr> calcola l'espressione <expr>\r\n");
       write(1,sbf,strlen(sbf));
    sprintf(sbf," end    uscita dalla chat\r\n");
       write(1,sbf,strlen(sbf));
    sprintf(sbf,"--- fine help ------------------\r\n");
       write(1,sbf,strlen(sbf));
}

int telfgets(char *nome,int n,int fhin,int fhot,long timeout,int riga, int passwd) {
  int nl = 0,skip=0,aspettonull = 0,ret,sret;
  long ct = 0L;
  unsigned char bf;
  fd_set fd;
  struct timeval timeo;

  if (riga) {
    if (n < 4)
      return -1;
    else
      n -= 4;
  }
  while(ct < timeout) {
    FD_ZERO(&fd);
    FD_SET(fhin,&fd);
    timeo.tv_sec  = 0L;
    timeo.tv_usec = 100000L;
    if ((sret = select(fhin+1, &fd, NULL, NULL, &timeo)) > 0) {
      if (FD_ISSET(fhin,&fd)) {
	if ((ret = read(fhin,&bf,1)) > 0) {
#ifdef DEBUG3
	  fputc(bf,fpdbg);
	  fflush(fpdbg);
#endif

	  if (aspettonull) {
#ifndef TELRAW
	    if (bf != '\0')
	      fprintf(fplog,"%ld: NON zero dopo \\r\n",rpid);
#else
	    if (bf != '\n')
	      fprintf(fplog,"%ld: NON lf dopo \\r\n",rpid);
#endif
	    nome[nl-1] = '\0';
	    write(fhot,"\r",1);
	    return 1;
	  }	    
	  if (skip > 0) {
	    skip--;
	    continue;
	  }
	  if (bf == 255) {
	    skip = 2;
	    continue;
	  }
	  if (bf == 0 || bf == '\n') 
	    continue;
	  if (bf == '\b' || bf == 127) {
	    if (nl > 0) {
	      write(fhot,"\b",1);
	      nl--;
	      continue;
	    }
	    else {
#ifdef DEBUG3
	  fprintf(fpdbg,"\nRENDE ZERO\n");
	  fflush(fpdbg);
#endif
	      return 0;
	    }
	  }
	  if (passwd)
	    write(fhot,"*",1);
	  else
	    write(fhot,&bf,1);

	  nome[nl++] = bf;
	  if(nome[nl-1] == '\r') {
	    /* PIPPO aspettonull = 1;
	    continue; */
	    nome[nl-1] = '\0';
	    write(fhot,"\r",1);
	    return 1;
	  }
	  if (nl == n) {
	    if (riga) {
	      nome[nl] =   ' ';
	      nome[nl+1] = '-';
	      nome[nl+2] = '>';
	      nome[nl+3] = '\0';
	    }
	    else
	      nome[nl] = '\0';
	    write(fhot,"\r",1);
	    return 1;
	  }
	}
	ct = 0;
	if (ret <= 0)
	  return (-1);
      }
    }
    else {
      ct++;    
#ifdef DEBUG2
      fprintf(fpdbg,"ct da gets %d\n",ct);
      fflush(fpdbg);
#endif
    }
  }
  nome[nl] = '\0';
  /* uscita per timeout */
  return 0;
}

int concurr_append(char *nomef,char *s) {
  FILE *fp;
  
  if ((fp = fopen(history,"a+")) == NULL)
    return -1;
  if (!filelock(F_WRLCK,fileno(fp),1,1,50))
    return 0; 
  fprintf(fp,"%s",s);
  if (!filelock(F_UNLCK,fileno(fp),1,1,50))
    return 0; 
  fclose(fp);
  
  return 1;
}

/* mode  :   F_WRLCK oppure F_UNLCK */
/* quanto:   in decimi di secondo */
int filelock(int mode,int fh,long dove,long lung,int quanto) {
  int er;
  long l;
  struct flock sflock;
  struct tms tm;
  struct timeval tv;
  
  if (quanto < 0)
    quanto = 0;
  if (fh < 0 || lung < 0L)
    return -1;
  sflock.l_type = mode;
  l = times(&tm);
  while (1) {
    sflock.l_type = mode;
    sflock.l_whence = 0;
    sflock.l_start = dove;
    sflock.l_len = lung;
    er = fcntl(fh,F_SETLK,&sflock);
    if (er != -1) 
      return 1;
    
    tv.tv_sec  = 0L;
    tv.tv_usec = 100000L;
    select(0, NULL, NULL, NULL, &tv);
    if ((times(&tm) -l) >  ((CLK_TCK/10) * quanto))
      return 0;
  }
}


long check_users(char *nome,long pid,long *lpid) {
  FILE *fp;
  char bf[256],*spid,*snome;
  long libero = -1,uguale = 0;
  int i;

  for( i=0 ; nome[i] != '\0' ; i++) {
    if (!isalnum(nome[i]) && nome[i] != ' ')
      return 0L;
  }
  if ((fp = fopen(fusers,"r+")) == NULL)
    if ((fp = fopen(fusers,"w+")) == NULL)
      return -1L;
  if (!filelock(F_WRLCK,fileno(fp),1,1,50))
    return 0L; 
  while(fread(bf,1,30,fp) > 0) {
    bf[30] = '\0';
#ifdef DEBUG2
    fprintf(fpdbg,"buf [%s",bf);
    fflush(fpdbg);
#endif
    if(!strcmp(bf,"                             \n")) {
      if(libero == -1)
	libero = ftell(fp) - 30;
      continue;
    }
    spid = strtok(bf,"|");
    snome = strtok(NULL,"|");
    /* controlla se i processi nell'elenco sono ancora attivi */
    if (kill((pid_t)atol(spid), 0) && errno == ESRCH) {
      fseek(fp,-30,SEEK_CUR);
      fflush(fp);
      if(libero == -1)
	libero = ftell(fp);
      fwrite("                             \n",1,30,fp);
      fflush(fp);
    }
    else {
      /* se il processo e' ancora attivo Compara il nome passato */
      if (!uguale && !strcmp(nome,snome))
	uguale = 1;
    }
  }
  if (uguale)
    return 0L;
  sprintf(bf,"%ld|%s|                               ",pid,nome);
  bf[29] = '\n';
  if (libero != -1)
    fseek(fp,libero,SEEK_SET);
  *lpid = ftell(fp);
  fwrite(bf,1,30,fp);
  fflush(fp);
  if (!filelock(F_UNLCK,fileno(fp),1,1,50))
    return 0L; 
  fclose(fp);
  
  return pid;
}



int remove_users(long lpid) {
  FILE *fp;
  
  if ((fp = fopen(fusers,"r+")) != NULL) {
    if (filelock(F_WRLCK,fileno(fp),1,1,50)) {
      if (!fseek(fp,lpid,SEEK_SET)) {
	fwrite("                             \n",1,30,fp);
	fflush(fp);
	if (filelock(F_UNLCK,fileno(fp),1,1,50)) {
	  fclose(fp);
	  return 1;
	}
      }
      filelock(F_UNLCK,fileno(fp),1,1,50);
    }
    fclose(fp);
  }

  return 0;
}


int display_users() {
  FILE *fp;
  char bf[256],bf2[256],*spid,*snome;
  int i = 0;
  
  if ((fp = fopen(fusers,"r+")) != NULL) {
    strcpy(bf2,"--- U T E N T I  C O N N E S S I ---\r\n");
    write(1,bf2,strlen(bf2));
    if (!filelock(F_WRLCK,fileno(fp),1,1,50))
      return 0; 
    while(fread(bf,1,30,fp) > 0) {
      bf[30] = '\0';
      if(!strcmp(bf,"                             \n"))
	continue;
      spid = strtok(bf,"|");
      snome = strtok(NULL,"|");
      /* controlla se i processi nell'elenco sono ancora attivi */
      if (kill((pid_t)atol(spid), 0) && errno == ESRCH) {
	fseek(fp,-30,SEEK_CUR);
	fflush(fp);
	fwrite("                             \n",1,30,fp);
	fflush(fp);
      }
      else {
	sprintf(bf2,"%12s",snome);
	if (i%6 == 5)
	  strcat(bf2,"\r\n");
	write(1,bf2,strlen(bf2));
	i++;
      }
    }
    strcpy(bf2,"\r\n------------------------------------\r\n");
    write(1,bf2,strlen(bf2));
  }
  else
    return 0;
  if (!filelock(F_UNLCK,fileno(fp),1,1,50))
    return 0; 
  fclose(fp);
  
  return 1;
}

void myexit(int ex,char *nome)
{
  char bf[64];

  if (lpid > -1)
    remove_users(lpid);
  if (nome) {
     sprintf(bf,"Uscita di %s\r\n",nome);
     if (concurr_append(history,bf) < 0 && ex == 0)
       ex = 1;
  }
#ifdef DEBUG
  if (fpdbg)
    fclose(fpdbg);
#endif
  if (fplog)
    fclose(fplog);
  exit(ex);
}












/*
 *   devil.c         chat server che si appoggia all'inetd per
 *   Vers. 0.9.5   l'aggancio delle connessioni telnet client.           
 *                 Copyright (C)  1999-2002  Matteo Nastasi 
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *     CONTACT:    nastasi@alternativeoutput.it
 *                 matteo.nastasi@milug.org
 *
 */
